/*****
*
* Copyright (C) 1998 - 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/time.h>
#include <unistd.h>

#include "list.h"
#include "common.h"
#include "print.h"
#include "plugin-common.h"
#include "plugin-common-prv.h"


typedef struct {
        struct list_head list;
        void *handle;

        struct list_head pclist;
        plugin_generic_t *plugin;
} plugin_entry_t;


static LIST_HEAD(all_plugin);
        

/*
 * Check if 'filename' look like a plugin.
 * FIXME : do real check.
 */
static int is_a_plugin(const char *filename) 
{
        return ( strstr(filename, ".so") ) ? 0 : -1;
}




/*
 * return a filename generated from 'dirname' & 'file'.
 */
static char *generate_filename(const char *dirname, const char *file) 
{
        char *filename;
        int len = strlen(dirname) + strlen(file) + 2;

        filename = malloc(len);
        if ( ! filename ) {
                err("malloc");
                return NULL;
        }
        
        snprintf(filename, len, "%s/%s", dirname, file);
        return filename;
}



/*
 *
 */
static plugin_entry_t *add_plugin_entry(void *handle, plugin_generic_t *p) 
{
        plugin_entry_t *pe;
        
        pe = malloc(sizeof(plugin_entry_t));
        if ( ! pe ) {
                err("malloc");
                return NULL;
        }
        
        pe->handle = handle;
        pe->plugin = p;
        
        INIT_LIST_HEAD(&pe->pclist);
        list_add_tail(&pe->list, &all_plugin);
        
        return pe;
}


/*
 *
 */
static plugin_container_t *create_container(plugin_entry_t *pe, plugin_generic_t *p) 
{
        plugin_container_t *pc;
        
        pc = calloc(1, sizeof(plugin_container_t));
        if ( ! pc ) {
                err("calloc");
                return NULL;
        }

        pc->plugin = p;
        pc->parent = pe;
        
        list_add_tail(&pc->int_list, &pe->pclist);
        
        return pc;
}


/*
 *
 */
static plugin_container_t *copy_container(plugin_container_t *pc) 
{
        plugin_entry_t *pe = (plugin_entry_t *) pc->parent;
        plugin_container_t *new;
        
        new = malloc(sizeof(plugin_container_t));
        if ( ! new ) {
                err("malloc");
                return NULL;
        }

        memcpy(new, pc, sizeof(plugin_container_t));
        new->already_used = 0;

        list_add_tail(&new->int_list, &pe->pclist);
        
        return new;
}


/*
 * Load a single plugin pointed to by 'filename'.
 */
static plugin_container_t *plugin_load_single(const char *filename) 
{
        void *handle;
        static int id = 0;
        plugin_generic_t *plugin;
        plugin_generic_t *(*init)(int id);
        plugin_entry_t *pe;
        
        handle = dlopen(filename, RTLD_NOW);
        if ( ! handle ) {
                err("can't open %s : %s", filename, dlerror());
                return NULL;
        }

        init = dlsym(handle, "plugin_init");
        if ( ! init ) {
                err("couldn't load %s : %s", filename, dlerror());
                return NULL;
        }

        plugin = init(id++);
        if ( ! plugin ) {
                err("There was an error initializing %s", filename);
                return NULL;
        }

        pe = add_plugin_entry(handle, plugin);
        if ( ! pe )
                return NULL;
         
        return create_container(pe, plugin);
}



/*
 *
 */
int plugin_load_from_dir(const char *dirname, int (*cb)(plugin_container_t *p)) 
{
        int ret;
        DIR *dir;
        char *filename;
        struct dirent *d;
        plugin_container_t *pc;
        
        dir = opendir(dirname);
        if ( ! dir ) {
                err("couldn't open directory %s", dirname);
                return -1;
        }

        while ( ( d = readdir(dir) ) ) {
                filename = generate_filename(dirname, d->d_name);

                ret = is_a_plugin(filename);
                if ( ret < 0 )
                        continue;
                
                pc = plugin_load_single(filename);
                if ( ! pc )
                        continue;
                
                if ( cb(pc) != 0 ) {
                        printf("Callback failed.\n");
                }
                
                        
                free(filename);
        }

        return 0;
}



/*
 *
 */
int plugin_register(plugin_container_t *pc, struct list_head *h, const char *infos)
{
        if ( pc->already_used++ && ! (pc = copy_container(pc)) ) {
                err("couldn't duplicate plugin container");
                return -1;
        }

        pc->infos = infos;
        
        list_add_tail(&pc->ext_list, h);
             
        return 0;
}


#if 0
/*
 *
 */
void plugin_run(plugin_container_t *pc, void *data) 
{
        struct timeval start, end;
        
        gettimeofday(&start, NULL);

        pc->plugin->run(data);
        
        gettimeofday(&end, NULL);
        
        pc->p_time += (double) end.tv_sec + (double) (end.tv_usec * 1e-6);
        pc->p_time -= (double) start.tv_sec + (double) (start.tv_usec * 1e-6);
        pc->p_count++;
        
}
#endif


/*
 *
 */
void plugin_print_stats(plugin_container_t *pc) 
{
        printf("%s ", pc->plugin->name);
        
        if ( pc->infos )
                p_print(stdout, "(infos=%s) :\n", pc->infos);
        else
                p_print(stdout, ":\n");

        p_print(stdout,  "\t\t- plugin: called %d time : %fs average\n",
                pc->p_count, pc->p_time / pc->p_count);

        p_print(stdout,  "\t\t- pfr: called %d time : %fs average\n",
                pc->pfr_count, pc->pfr_time / pc->p_count);
}



/*
 *
 */
void plugins_print_stats(void) 
{
        plugin_container_t *pc;
        plugin_entry_t *pe;
        struct list_head *tmp, *tmp2;
        
        list_for_each(tmp, &all_plugin) {
                pe = list_entry(tmp, plugin_entry_t, list);
                
                list_for_each(tmp2, &pe->pclist) {
                        pc = list_entry(tmp2, plugin_container_t, int_list);
                        plugin_print_stats(pc);
                }
        }               
}


/*
 *
 */
void plugins_print_opts(void) 
{
        plugin_entry_t *pe;
        struct list_head *tmp;
        
        list_for_each(tmp, &all_plugin) {
                pe = list_entry(tmp, plugin_entry_t, list);

                if ( pe->plugin->opt ) {
                        fprintf(stderr, "\nUsage for %s plugin "
                                "(use \"-m %s\" to supply option) :\n",
                                pe->plugin->name, pe->plugin->name);

                        pe->plugin->opt(0, NULL);
                }
        }
}



/*
 *
 */
#include <getopt.h>

static int _plugin_set_opts(plugin_entry_t *pe, int argc, char **argv) 
{
        int i, ret;
        char *nargv[argc];
        int s_optind;
        
        if ( ! pe->plugin->opt ) {
                err("Plugin %s can't accept option", pe->plugin->name);
                return -1;
        }
        
        nargv[0] = pe->plugin->name;
        
        for (i = 1; i < argc && argv[i][1] != 'm'; i++) 
                nargv[i] = argv[i];

        s_optind = optind;
        optind = 0;        
        ret = pe->plugin->opt(i, nargv);
        optind = (s_optind + (optind - 1));
        
        return ret;
}



/*
 *
 */
int plugin_set_opts(char *pname, int argc, char **argv) 
{
        int ret;
        plugin_entry_t *pe;
        struct list_head *tmp;

        list_for_each(tmp, &all_plugin) {
                pe = list_entry(tmp, plugin_entry_t, list);

                ret = strcmp(pe->plugin->name, pname);
                if ( ret == 0 ) {
                        return _plugin_set_opts(pe, argc, argv);
                }
        }
        
        err("Couldn't find plugin %s", pname);

        return -1;
}
