#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "packet.h"
#include "match.h"

static void *_compile_proto(const char *proto);

static struct _rtbl {
        const char *rule;
        unsigned int id;
        void *(*compile_value)(const char *value);
        void *data;
} rtbl[] = {
        { "proto", 0, _compile_proto },
        { "sip", 1, NULL        },
        { "dip", 2, NULL        },
        { "sport", 3, NULL      },
        { "dport", 4, NULL      },
        { "tcp_flag", 5, NULL   },
        { "ip_flag",  6, NULL   },
        { "content", 7, NULL    },
        { "hexcontent", 8, NULL },
        { "msg", 9, NULL        },
        { NULL, 10              },
};



static struct _match *tree_beg = NULL, *tree_cur_child, *tree_cur;

/*
 * Compile func.
 */
static void *_compile_proto(const char *proto) 
{
        int ret;

        ret = strcmp(proto, "tcp");
        if (ret == 0)
                return (void *) IPPROTO_TCP;

        ret = strcmp(proto, "udp");
        if (ret == 0)
                return (void *) IPPROTO_UDP;

        return NULL;
}


static struct _match *node_alloc(void) 
{
        struct _match *ptr;
        
        ptr = calloc(1, sizeof(struct _match));
        if ( ! ptr ) 
                fprintf(stderr, "memory exhausted.\n");

        return ptr;
}


static int compare(struct _match *root, struct _match *rule) 
{
        int ret;
        
        while ( root->next ) {
                printf("test\n");
                if ( ! (root->id == rule->id && root->data == rule->data) ) {
                        root->next = rule->next;
                        return;
                }

                return compare(root->child, rule->child);
                
                root = root->next;
        }
        
        return -1;
}



static int append_tree(struct _match *rule) 
{
        int ret;
        
        if (!tree_beg) 
                tree_cur = tree_beg = rule;
        else {
                ret = compare(tree_beg, rule);
        }
}



static int append_single_rule(struct _match **tmp, struct _rtbl *rptr, const char *value)
{
        struct _match *new;
        static struct _match *cur = NULL;
        
        new = node_alloc();
        if (! new )
                return -1;

        new->debug = (char *) strdup(rptr->rule);
        if (! new->debug )
                return -1;

        new->data = rptr->data;
        new->id = rptr->id;
                
        if ( ! *tmp )
                cur = *tmp = new;
        else {
                cur->child = new;
                cur = cur->child;
        }
                        
        return 0;
}


static struct _rtbl *lookup_rule(const char *rule, const char *value) 
{
        int i;

        for ( i = 0; rtbl[i].rule; i++ )
                if ( strcmp(rtbl[i].rule, rule) == 0 ) {
                        if ( rtbl[i].compile_value ) 
                                rtbl[i].data = rtbl[i].compile_value(value);
                        
                        return &rtbl[i];
                }
        

        return NULL;
}



/*
 * syntax to parse is "rule:whatever-content;"
 */
static int get_next_rule(char **line, int lino, char **rule, char **value) 
{
        if ( *line == NULL )
                return -1;

        while (**line == ' ')
                *line = *line + 1;

        if ( **line == 0 )
                return -1;
        
        *rule = strsep(line, ":");
        if ( ! *rule ) {
                fprintf(stderr, "line %d: the rule file doesn't seem valid.\n", lino);
                return -1;
        }
                
        *value = strsep(line, ";");
        if ( ! *value ) {
                fprintf(stderr, "line %d: couldn't get value for rule %s.\n", lino, *rule);
                return -1;
        }
        
        return 0;
}

        
        
static void parse_line(char *line, int lino) 
{
        int ret;
        char *rule, *value;
        struct _rtbl *ptr;
        struct _match *tmp = NULL;

        while ( get_next_rule(&line, lino, &rule, &value) == 0 ) {
                
                ptr = lookup_rule(rule, value);
                if ( ! ptr ) {
                        fprintf(stderr, "couldn't find handler for rule %s\n", rule);
                        return;
                }

                ret = append_single_rule(&tmp, ptr, value);
                if ( ret < 0 )
                        return;
        }

        if ( tmp ) {
                ret = append_tree(tmp);
                if ( ret < 0 )
                        return;
        }
}

static void print_space(int num) 
{
        while (num--)
                putchar('\t');
}


static void dump_tree(struct _match *root, int depth) 
{
        int ret;
        
        while ( root ) {
                print_space(depth);
                printf("- depth=%d, debug=\"%s\", data = %p\n", depth, root->debug, root->data);

                dump_tree(root->child, depth + 1);
                root = root->next;
        }
}


int parse_rule_file(const char *filename) 
{
        int i;
        FILE *fd;
        char buf[1024], *p;

        fd = fopen(filename, "r");
        if ( ! fd ) {
                fprintf(stderr, "coudln't open %s : %s\n", filename, strerror(errno));
                return -1;
        }

        i = 0;
        while ( fgets(buf, sizeof(buf), fd) ) {
                i++;
                
                p = buf;
                while (*p == ' ' || *p == 0 ) {
                        if ( *p == 0 )
                                break;
                        p++;
                }   

                if ( *p == '#' )
                        break;

                p[strlen(p) - 1] = 0;
                parse_line(buf, i);
        }
        

        fclose(fd);

        printf("\n*** dumping match tree ***\n");
        dump_tree(tree_beg, 0);
        printf("*** done ***\n");
        
        return 0;
}
