/*****
*
* Copyright (C) 1998 - 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <pcap.h>

#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

#include "pconfig.h"
#include "config_devices.h"
#include "capture.h"
#include "packet-decode.h"
#include "timer.h"
#include "common.h"
#include "print.h"
#include "packet.h"

void __libprelude_FreeCachedData(void);

static pcap_t *pd_file = NULL;
static pcap_dumper_t *p_dump = NULL;
static unsigned long long packet_counter = 0;
static fd_set fds;

extern Pconfig_t config;
extern Device_t **devices;
extern unsigned long prelude_reporting_time;


/*
 * Search for the datalink handler corresponding
 * to the datalink of type 'type'.
 *
 * Return NULL if no handler was found.
 */
static pcap_handler search_datalink_handler(int type)
{
        struct datalink_handler {
                pcap_handler handler;
                int type;
        } p[] = {
                {	SliceAndStoreEtherPkt, DLT_EN10MB	},
                {	SliceAndStoreEtherPkt, DLT_IEEE802	},
                {       SliceAndStoreNullPkt,  DLT_NULL         },
                {       SliceAndStorePppPkt,   DLT_PPP          },
                {       SliceAndStorePppBsdosPkt, DLT_PPP_BSDOS },
                {       SliceAndStoreFddiPkt, DLT_FDDI          },
                {       SliceAndStoreAtmPkt, DLT_ATM_RFC1483    },
                {       NULL, 0                                 }, 
        };

        while (p != NULL) {
                if ( type == p->type )
                        return p->handler;
        }

        err("Unknow datalink type 0x%x.\n", type);
        
        return NULL;
        
}



/*
 * Setup bpf rule 'bpf' for packet device 'pd',
 * with netmask 'netmask'.
 */
static int setup_bpf(pcap_t *pd, char *bpf, bpf_u_int32 *netmask)
{
	struct bpf_program fcode;
        int ret;

        if ( bpf == NULL )
		return 0;

        ret = pcap_compile(pd, &fcode, bpf, 1, *netmask);
	if ( ret < 0 ) {
		fprintf(stderr, "pcap: %s.\n", pcap_geterr(pd));
		exit(1);
	}

	ret = pcap_setfilter(pd, &fcode);
        if ( ret < 0 ) {
		fprintf(stderr, "pcap: %s.\n", pcap_geterr(pd));
		exit(1);
	}
        
	return 0;
}



/*
 * Setup pcap to capture from file 'filename', with bpf rule 'bpf'.
 */
static int setup_capture_from_file(const char *filename, char *bpf)
{
	char error[PCAP_ERRBUF_SIZE];
	bpf_u_int32 netmask = 0;

	p_print(stdout, "\tReading packet from file : %s.\n", filename);

	pd_file = pcap_open_offline(filename, error);
	if (pd_file == NULL) {
		err("pcap_open_offline: %s", error);
		return -1;
	}

	return setup_bpf(pd_file, bpf, &netmask);
}



/*
 * Setup pcap to capture device 'pd' to a file 'filename'.
 */
static pcap_handler setup_capture_to_file(const char *filename, pcap_t *pd) 
{
        p_print(stdout, "Dumping packet to %s.\n", filename);

        p_dump = pcap_dump_open(pd, filename);
        if (! p_dump) {
                err("pcap_dump_open: %s", pcap_geterr(pd));
                return NULL;
        }

        return pcap_dump;
}



/*
 * Setup pcap to capture from device 'dev'.
 */
static int setup_capture_from_device(Device_t *dev)
{
	char err[PCAP_ERRBUF_SIZE];
	bpf_u_int32 localnet, netmask;
        int ret;
        int snaplen;
        
	p_print(stdout, "\tReading packet from device : %s.\n", dev->name);

        snaplen = ( config.snaplen > 0 ) ? config.snaplen : 2000;
        decode_init(1000, snaplen);
        
	dev->pd = pcap_open_live(dev->name, snaplen, 1, 1000, err);
	if (dev->pd == NULL) {
		err("pcap_open_live: %s", err);
                return -1;
	}

        ret = pcap_lookupnet(dev->name, &localnet, &netmask, err);
	if ( ret < 0) {
		pcap_close(dev->pd);
		err("pcap_lookupnet: %s", err);
                return -1;
	}

	dev->p_handler = search_datalink_handler(pcap_datalink(dev->pd));
        if ( ! dev->p_handler )
                return -1;
        
	dev->fd = pcap_fileno(dev->pd);

	return setup_bpf(dev->pd, dev->bpf_rule, &netmask);
}



/*
 * Setup pcap for capture.
 */
static int setup_capture(void) 
{
        int i, ret;

        if ( config.read_pkts_file != NULL )
                return setup_capture_from_file(config.read_pkts_file,
                                               config.read_pkts_bpf);
        
        if ( config.interface_nbr > 1 )
                FD_ZERO(&fds);

	for (i = 0; i < config.interface_nbr; i++) {
		ret = setup_capture_from_device(devices[i]);
                if ( ret < 0 )
                        return -1;
		/*
		 * fd is a copy of the file descriptor used by pcap
                 * This is only used when looping throught many interface.
		 */
                if ( config.interface_nbr > 1 ) 
                        FD_SET(devices[i]->fd, &fds);
                
	 	/*
		 * If the user want it,
		 * we configure this interface to be recorded directly to a file.
		 */
		if (config.record_pkts != NULL)
                        devices[i]->p_handler = setup_capture_to_file(config.record_pkts, devices[i]->pd);
	}

        return 0;
}



/*
 * Start capturing from multiple devices,
 * only called if more than 1 capture device were configured.
 */
static int capture_from_multiple_devices(void) 
{
        int ret;
	unsigned int i = 0;
	fd_set fds_bkp;
        struct timeval tv;
        unsigned long last = 0;

        memcpy(&fds_bkp, &fds, sizeof(fd_set));
        
        while (1) {
                tv.tv_usec = 0;
                tv.tv_sec = 1;
                
		ret = select(devices[config.interface_nbr - 1]->fd + 1, &fds, NULL, NULL, &tv);
                if ( ret == 0 ) {
                        /*
                         * Timeout occured, wake up timer.
                         */
                        wake_up_timer();
                        memcpy(&fds, &fds_bkp, sizeof(fds));
                        continue;
                }

		for (i = 0; i < config.interface_nbr; i++) {

                        ret = FD_ISSET(devices[i]->fd, &fds);
                        if ( ret != -1 ) {
				packet_counter++;
                                                                 
				pcap_dispatch(devices[i]->pd, 1,
					      devices[i]->p_handler, NULL);
                                
                                gettimeofday(&tv, NULL);
                                if ( (tv.tv_sec - last) >= 1 ) {
                                        last = tv.tv_sec;
                                        wake_up_timer();
                                }
                                
                                __libprelude_FreeCachedData();
                                
			}
		}
                memcpy(&fds, &fds_bkp, sizeof(fds));
	}
}



/*
 * Start capturing from a single device.
 */
static int capture_from_single_device(void) 
{
        int ret;
        struct timeval tv;
        unsigned long last = 0;
                
        while ( (ret = pcap_read(devices[0]->pd, 1, devices[0]->p_handler, NULL)) >= 0) {
                packet_counter++;

                gettimeofday(&tv, NULL);
                if ( (tv.tv_sec - last) >= 1 ) {
                        last = tv.tv_sec;
                        wake_up_timer();
                }

                __libprelude_FreeCachedData();
        }

        err("pcap_loop: %s", pcap_geterr(devices[0]->pd));

        return ret;
}



/*
 * Start capturing from a file.
 */
static int capture_from_file(void) 
{
        int ret;
        pcap_handler p_handler;
        
        /*
         * We configure pcap for this operation.
         */
        p_handler = search_datalink_handler(pcap_datalink(pd_file));
        if ( ! p_handler )
                return -1;
        /*
         * We read the packet from the file, incrementing the packet counter
         * of one for each packet.
         */
        while ( (ret = pcap_dispatch(pd_file, 1, p_handler, NULL)) > 0)
                packet_counter++;

        if ( ret < 0 )
                err("pcap_dispatch: %s", pcap_geterr(pd_file));
             
        return ret;
}



/*
 * Gather stats from device 'pd'.
 */
static void pcap_dump_stat_from_device(pcap_t *pd) 
{
        int ret;
        struct pcap_stat stat;

        ret = pcap_stats(pd, &stat);
        if ( ret < 0 ) {
                err("pcap_stats: %s.", pcap_geterr(pd));
                return;
        }
        
        printf("%u packets dropped by the kernel.\n\n", stat.ps_drop);
}



/*
 * Dump pcap stats.
 */
static void dump_pcap_stats() 
{
        int i;
        
        for (i = 0; i < config.interface_nbr; i++)
                pcap_dump_stat_from_device(devices[i]->pd);
        
	if (pd_file != NULL)
                pcap_dump_stat_from_device(pd_file);
}



/*
 * Dump some process stats, gathered using getrusage.
 */
static void dump_system_stats(void) 
{
        struct rusage rusage;
	double total_time;
        
        getrusage(RUSAGE_SELF, &rusage);

        total_time = ((double) rusage.ru_utime.tv_usec +
                      (double) rusage.ru_stime.tv_usec) * 1e-6;
	total_time += ((double) rusage.ru_utime.tv_sec +
                       (double) rusage.ru_stime.tv_sec);
        
	printf("Average cpu time by packet : %fs, %fms, %fus.\n",
	       (total_time / (double) packet_counter),
               (total_time / (double) packet_counter) * 1e3,
	       (total_time / (double) packet_counter) * 1e6);
                       
        printf("\nPage reclaims = %ld\n", rusage.ru_minflt);
        printf("Page faults = %ld\n", rusage.ru_majflt);
        printf("Swap = %ld\n", rusage.ru_nswap);
}



/*
 * Print some gathered statistics.
 */
void capture_stats(void)
{
        printf("%qd packets received by filter. (prelude counted),"
               " will reset after 2e64-1.\n", packet_counter);
        
        dump_pcap_stats();
        dump_system_stats();
}



/*
 * Setup the capture device[s],
 * then call the wanted capture function.
 */
int capture_start(void)
{
	int ret;
            
        ret = setup_capture();
        if ( ret < 0 )
                return -1;
        
	if ( config.read_pkts_file != NULL ) 
                return capture_from_file();

        if ( config.interface_nbr > 1 )
                return capture_from_multiple_devices();
        else 
                return capture_from_single_device();

        return -1;
}



/*
 * Stop packet capture.
 */
void capture_stop(void)
{        
	if ( config.interface_nbr > 0 ) {
                int i;
                
                for (i = 0; i < config.interface_nbr; i++)
			pcap_close(devices[i]->pd);
	}
        
	if ( config.read_pkts_file != NULL )
		pcap_close(pd_file);
                
	if ( p_dump != NULL )
		pcap_dump_close(p_dump);
}




