/*****
*
* Copyright (C) 1999, 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <regex.h>

#include "packet.h"
#include "detect_plugins.h"
#include "filter_funcs.h"
#include "common.h"
#include "lookup.h"
#include "pluginlib.h"


extern unsigned long int prelude_resolve(const char *host);

extern struct ether_header __ether_hdr;
extern struct ip           __ip_hdr;
extern struct __iphdr      __ip;
extern struct tcphdr       __tcp_hdr;
extern struct __tcphdr     __tcp;
extern struct udphdr       __udp_hdr;
extern struct icmp         __icmp_hdr;
extern struct igmp         __igmp_hdr;
extern struct arphdr       __arp_hdr;
extern PktData_t           __pkt_data;
extern Packet_t packet[MAX_PKTDEPTH];


int verify_shost(unsigned long int arg) 
{
        if ( arg == __ip_hdr.ip_src.s_addr )
                return 0;
        return -1;
}

int verify_dhost(unsigned long int arg) 
{
        if ( arg == __ip_hdr.ip_dst.s_addr )
                return 0;
        return -1;
}

int verify_sport(unsigned long int arg) 
{
        if ( __ip_hdr.ip_p & IPPROTO_TCP ) 
                if ( arg == __tcp_hdr.th_sport )
                        return 0;
        
        if ( __ip_hdr.ip_p & IPPROTO_IP )
                if ( arg == __udp_hdr.uh_sport )
                        return 0;
        return -1;
}

int verify_dport(unsigned long int arg) 
{
        
        if ( __ip_hdr.ip_p & IPPROTO_TCP ) 
                if ( arg == __tcp_hdr.th_dport )
                        return 0;
        
        if ( __ip_hdr.ip_p & IPPROTO_IP )
                if ( arg == __udp_hdr.uh_dport )
                        return 0;        
        return -1;
}

int verify_proto(unsigned long int arg) 
{
        int ret;
        __u8 p = (__u8) arg;
        
        ret = prelude_GetDepth(packet, p_ip, depth_ip);
        if ( ret < 0 )
                return -1;

        if ( __ip_hdr.ip_p & p )
                return 0;
        
        return -1;
}

int verify_tcpflag(unsigned long int arg) 
{
        int *flag = (int *) arg;
        int i;

        for ( i = 0; flag[i] != -1; i++ ) {
                if ( ! (__tcp_hdr.th_flags & flag[i]) )
                        return -1;
        }
        
        return 0;
}

int verify_ipfrag(unsigned long int arg)
{
	unsigned int off;

	off = ntohs(__ip_hdr.ip_off);
        if ( off & 0x3fff )
		return 0;
	
	return -1;
}

int verify_tcpoob(unsigned long int arg) 
{
        /*
         * Is this right ?
         */
        
        if ( __tcp_hdr.th_urp )
                return 0;
        return -1;
}

/*
 * Using regex here is probably overkill.
 */
int verify_regcontent(unsigned long int arg) 
{
        if ( __pkt_data.size > 0 )
                return regexec((regex_t *)arg, __pkt_data.data, 0, NULL, 0);

        return -1;
}

/*
 * Strstr seem to be more powefull than all
 * the implementation of the Boyer Moore algorithm
 * i have tryed. strstr beat them all.
 * ( probably because it is asm optimized,
 * and that is use this algorithm itself.
 */
int verify_content(unsigned long int arg) 
{
        char *pattern = (unsigned char *) arg;

        if ( __pkt_data.size > 0 )
                if ( strstr(__pkt_data.data, pattern) )
                        return 0;
        
        return -1;
}

int verify_tcpopt(unsigned long int arg) 
{
        int *opts = (int *) arg;
        int i, ret;
        pktopt_t option;

        prelude_InitTcpOptParsing(&option, &__tcp);
        
        for ( i = 0; opts[i] != -1; i++ ) {
                ret = prelude_SearchTcpOption(&option, opts[i]);
                if ( ret != 0 )
                        return -1;
        }
                
        return 0;
}

int verify_ipopt(unsigned long int arg) 
{
        int *opts = (int *) arg;
        int i, ret;
        pktopt_t option;

        prelude_InitIpOptParsing(&option, &__ip);
        
        for ( i = 0; opts[i] != -1; i++ ) {
                ret = prelude_SearchIpOption(&option, opts[i]);
                if ( ret != 0 )
                        return -1;
        }
        
        return 0;      
}

extern struct value_name tcpflags_table;

unsigned long int compile_tcpflag(const char *arg) 
{
        int *flags = NULL;
        char *ret, *p, *ptr;
        int n = 0;
        int f;
        
        p = ptr = strdup(arg);
        while ( ptr ) {
                ret = strsep(&ptr, ",");
                f = lookup_val(&tcpflags_table, ret);
                if ( f == -1 )
                        continue;
                
                flags = realloc(flags, (n + 1) * sizeof(int));
                flags[n] = f;
                n++;
        }

        flags = realloc(flags, (n + 1) * sizeof(int));
        flags[n] = -1;

        free(p);
        
        return (unsigned long int) flags;
}

extern struct value_name ipproto_table;

unsigned long int compile_proto(const char *arg)
{
        return lookup_val(&ipproto_table, arg);
}

unsigned long int compile_port(const char *arg) 
{
        return(htons(atoi(arg)));
}

unsigned long int compile_host(const char *arg) 
{
        return(prelude_resolve(arg));
}

static regex_t *create_regex(const unsigned char *pattern) 
{
        char err[1024];
        regex_t *reg;
        int ret;
        
        reg = (regex_t *) malloc(sizeof(regex_t));
        if ( ! reg ) {
                err("malloc(%d)", sizeof(regex_t));
                return NULL;
        }
        
        ret = regcomp(reg, pattern, REG_NOSUB);
        if ( ret < 0 ) {
                regerror(ret, reg, err, sizeof(err));
                return NULL;
        }

        return reg;
}


unsigned long int compile_hexcontent(const char *arg) 
{
        unsigned char *byte_string = NULL;
        char *p, *ptr, *ret;
        size_t n = 0;
        regex_t *regex;
        
        p = ptr = strdup(arg);
                    
        while ( ptr ) {
                ret = strsep(&ptr, " ");
                byte_string = realloc(byte_string, (n + 1) * sizeof(unsigned char));
                byte_string[n] = (unsigned char) strtoul(ret, NULL, 16);
                n++;
        }
        
        byte_string = realloc(byte_string, (n + 1) * sizeof(unsigned char));
        byte_string[n] = '\0';        

        free(p);

        regex = create_regex(byte_string);
        if ( ! regex )
                return -1;

        return (unsigned long int) regex;
}

unsigned long int compile_regcontent(const char *arg) 
{
        regex_t *regex;
        
        regex = create_regex(arg);
        if ( ! regex )
                return -1;

        return (unsigned long int) regex;
}

        

unsigned long int compile_content(const char *arg) 
{
        return ( unsigned long int ) arg;
}


static unsigned long int generic_opt(struct value_name *table, const char *arg) 
{
        char *r, *p, *ptr;
        int n = 0, value, *opts = NULL;
            
        p = ptr = strdup(arg);
        while ( ptr ) {
                r = strsep(&ptr, ",");
                value = lookup_val(table, r);
                if ( value == -1 )
                        continue;
                opts = realloc(opts, (n + 1) * sizeof(int));
                opts[n] = value;
                n++;
        }

        opts = realloc(opts, (n + 1) * sizeof(int));
        opts[n] = -1;

        free(p);
        
        return (unsigned long int) opts;     
}

extern struct value_name tcpopts_table;

unsigned long int compile_tcpopt(const char *arg) 
{
        return generic_opt(&tcpopts_table, arg);
}

extern struct value_name ipopts_table;

unsigned long int compile_ipopt(const char *arg) 
{
        return generic_opt(&ipopts_table, arg);    
}













