/*****
*
* Copyright (C) 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <sys/time.h>

#include "common.h"
#include "hostdb.h"

#define HASH_SIZE 1024

#define hkey(saddr, daddr, prot) \
               (( (saddr) ^ (daddr) ^ (prot)) & (HASH_SIZE - 1))
#define host_key(ip) hkey(ip->ip_src.s_addr, ip->ip_dst.s_addr, ip->ip_p)

extern int plugins_id_max;
static hostdb_t *host_hash[HASH_SIZE];




/*
 * Compare two IP address using the hash criteria.
 */
static int ip_match(const struct ip *ip1, const struct ip *ip2) 
{
        return ( ip1->ip_src.s_addr == ip2->ip_src.s_addr &&
                 ip1->ip_dst.s_addr == ip2->ip_dst.s_addr &&
                 ip1->ip_p == ip2->ip_p );
}



/*
 * Loop all hostdb entry from bucket 'bucket',
 * and call ip_match() to see if the wanted IP is part of this entry.
 */
static hostdb_t *search(hostdb_t *bucket, const struct ip *ip) 
{
        while ( bucket != NULL ) {
                if ( ip_match(bucket->ip, ip) )
                        return bucket;
                bucket = bucket->next;
        }        
        return NULL;
}



/*
 * Add an hostdb entry to the hash table.
 */
static void host_add(hostdb_t **bucket, hostdb_t *h) 
{
        h->next = *bucket;
        if ( h->next )
                h->next->prev = h;
        *bucket = h;
}



/*
 * Delete an hostdb entry from the hash table. 
 */
static void host_del(hostdb_t **bucket, hostdb_t *h) 
{        
        if ( h->next )
                h->next->prev = h->prev;
        if ( h->prev )
                h->prev->next = h->next;
        else *bucket = h->next;

        free(h->pdata);
        free(h);
}




/*
 * Create a new hostdb entry.
 */
hostdb_t *hostdb_new(const struct ip *ip) 
{
        int key = host_key(ip);
        hostdb_t *new;
        
        new = malloc(sizeof(hostdb_t));
        if (! new ) {
                err("malloc(%d)", sizeof(hostdb_t));
                return NULL;
        }

        new->pdata = calloc(1, plugins_id_max * sizeof(unsigned long *));
        if ( ! new->pdata ) {
                err("malloc(%d):",  plugins_id_max * sizeof(unsigned long *));
                return NULL;
        }
        
        new->prev = NULL;
        new->refcount = 0;
        new->ip = ip;
        new->key_cache = key;
        
        host_add(&host_hash[key], new);

        return new;
}



/*
 * Delete an hostdb entry
 */
void hostdb_del(hostdb_t *h, const unsigned int pid)
{
        h->pdata[pid] = 0;

        if ( --h->refcount == 0 ) 
                host_del(&host_hash[h->key_cache], h);
}



/*
 * Search for an host in hostdb.
 */
hostdb_t *hostdb_search(const struct ip *ip) 
{
        return search(host_hash[host_key(ip)], ip);
}



/*
 * Init the hostdb modules, we just 0 our hash.
 */
int hostdb_init(void) 
{
        memset(host_hash, 0, HASH_SIZE - 1);
        
        return 0;
}
















