/*****
*
* Copyright (C) 1998, 1999, 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <pcap.h>

#include "list.h"
#include "packet.h"
#include "plugin-common.h"
#include "plugin-common-prv.h"
#include "detect-plugins.h"
#include "config-engine.h"
#include "pconfig.h"
#include "config_devices.h"
#include "auth.h"
#include "common.h"

#define FILE CONFIG_DIR"/prelude.conf"

extern Pconfig_t config;
int config_quiet;
Device_t **devices = NULL;


static void print_usage(int argc, char **argv)
{
	fprintf(stderr, "Usage :\n\n");
	fprintf(stderr, "\t-d --daemonize Daemon mode.\n");
        fprintf(stderr, "\t-q --quiet Quiet mode.\n");
        fprintf(stderr, "\t-u --user create new Prelude user.\n");
        fprintf(stderr, "\t-s --snaplen specify the packet snaplen.\n");
        fprintf(stderr, "\t-r --read-from <filename> Read file previously recorded.\n");
	fprintf(stderr, "\t-w --write-to <filename> Record packets to specified file. Don't analyze them.\n");
	fprintf(stderr, "\t-a --report-addr <server> Specify a report server.\n");
	fprintf(stderr, "\t-p --report-port <port> Specify the report server port.\n");
	fprintf(stderr, "\t-i <interface> Specify an interface to read packet from,\n");
        fprintf(stderr, "\t\t\teventually followed by a quoted bpf rules.\n\n");

        fprintf(stderr, "Usage (plugin help):\n\n");
        fprintf(stderr, "\t-m --plugin <name> <option> to set/get plugin specific options.\n");

        plugins_print_opts();
}

static void configure_address(config_t *cfg) 
{
        const char *ret;
        
        if ( config.addr )
                return;
        
        ret = config_get(cfg, "Prelude", "report addr");
        if ( !ret ) {
                fprintf(stderr, "No report server specified in config file or command line.\n");
                exit(1);
        }
        
        config.addr = ret;
}

static void configure_port(config_t *cfg) 
{
        const char *ret;
        
        if ( config.port != 0 )
                return;
        
        ret = config_get(cfg, "Prelude", "report port");
        if ( !ret ) {
                fprintf(stderr, "No report server port specified in config file or command line.\n");
                return;
                exit(1);
        }

        config.port = atoi(ret);
}


static int verify_opt(config_t *cfg)
{
	if (config.interface_nbr == 0 && config.read_pkts_file == NULL) {
		fprintf(stderr, "\nNo interface nor file to read the packet from were specified.\n\n");
		print_usage(0, NULL);
		return -1;
	}

	if (config.read_pkts_file != NULL && config.interface_nbr > 0) {
		fprintf(stderr, "\nCan't read packet from both interface and file.\n\n");
		return -1;
	}

	if (config.read_pkts_file != NULL && config.record_pkts != NULL) {
		fprintf(stderr, "\nCan't both read and write packet from a file.\n\n");
		return -1;
	}

        configure_address(cfg);
        
        if ( strcmp(config.addr, "unix") != 0 ) 
                configure_port(cfg);
        
	return 0;
}

static int alloc_interface(char *arg)
{
	if (arg == NULL) {
		fprintf(stderr, "\nOption -i take an argument.\n\n");
		return -1;
	}

	config.interface_nbr++;
	if (devices == NULL) {
		devices = malloc(config.interface_nbr * sizeof(Device_t *));
	} else {
		devices = realloc(devices, config.interface_nbr * sizeof(Device_t *));
	}

	if (devices == NULL) {
		fprintf(stderr, "realloc: %s.\n", strerror(errno));
		return -1;
	}
	devices[config.interface_nbr - 1] = malloc(sizeof(Device_t));
        devices[config.interface_nbr - 1]->name = arg;
	devices[config.interface_nbr - 1]->bpf_rule = NULL;

	return 0;
}


int pconfig_set(int argc, char **argv)
{
        config_t *cfg;
        struct option opts[] = 
        {
                { "help", no_argument, NULL, 'h' },
                { "user", no_argument, NULL, 'u' },
                { "quiet", no_argument, NULL, 'q' },
                { "daemonize", no_argument, NULL, 'd' },
                { "snaplen", required_argument, NULL, 's' },
                { "write-to", required_argument, NULL, 'o' },
                { "read-from", required_argument, NULL, 'r' },
                { "report-addr", required_argument, NULL, 'a' },
                { "report-port", required_argument, NULL, 'p' },
                { "retry-report", required_argument, NULL, 't' },
                { "interface", required_argument, NULL, 'i' },
                { 0, 0, 0, 0 }
        };                              


        /* Default */
	config_quiet = 0;
	config.daemonize = 0;
        config.snaplen = -1;
	config.record_pkts = NULL;
	config.read_pkts_file = NULL;
	config.interface_nbr = 0;
	config.read_pkts_bpf = NULL;
	config.addr = NULL;
	config.port = 0;
	/*         */

	while (1) {
		int c;

                c = getopt_long(argc, argv, "t:i:qw:r:dhus:p:m:", opts, NULL);
		if (c == -1)
			break;

		switch (c) {

                case 'u':
                        exit(auth_create_account());

                case 'q':
			config_quiet = 1;
			break;

                case 'd':
			config.daemonize = 1;
			break;

                case 's':
                        config.snaplen = atoi(optarg);
                        break;

                case 'w':
			config.record_pkts = optarg;
			break;

                case 'r':
			config.read_pkts_file = optarg;
			if (argv[optind] != NULL && *argv[optind] != '-') {
				config.devices[config.interface_nbr - 1]->bpf_rule = argv[optind];
				optind++;
			}
			break;

                case 'i':
			if (alloc_interface(optarg) == -1)
				return -1;
                        
			if (argv[optind] != NULL && *argv[optind] != '-') {
				devices[config.interface_nbr - 1]->bpf_rule = argv[optind];
				optind++;
			}
			break;

                case 'a':
			config.addr = optarg;
			break;
                        
		case 'p':
			config.port = atoi(optarg);
                        break;

                case 'm':
                        if ( plugin_set_opts(optarg, argc - (optind - 1), &argv[optind - 1]) < 0 )
                                exit(1);
                        break;
                        
		case 'h':
			print_usage(argc, argv);
			exit(0);

                default:
		}
	}

        cfg = config_open(FILE);
        if ( ! cfg ) {
                err("opening configuration file %s", FILE);
                return -1;
        }
        
	if (verify_opt(cfg) == -1)
		return -1;
        
	return 0;
}
