/*****
*
* Copyright (C) 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "list.h"
#include "plugin_filters.h"
#include "filter_funcs.h"
#include "common.h"

#define skip_space(rule) while ( rule[i] == ' ' ) i++;
#define skip_equal(rule) while ( rule[i] == '=' ) i++;

struct bpf_rule {
        const char *rule_name;
        unsigned long int (*compile)(const char *arg);
        int (*run)(unsigned long int arg);
};

static struct bpf_rule rules_table[] = {
        { "proto", compile_proto, verify_proto },
        { "dst_host", compile_host, verify_dhost },
        { "src_host", compile_host, verify_shost },
        { "src_port", compile_port, verify_sport },
        { "dst_port", compile_port, verify_dport },
        { "tcp_flag", compile_tcpflag, verify_tcpflag },
        { "tcp_oob", NULL, verify_tcpoob              },
	{ "ipfrag", NULL, verify_ipfrag               },
        { "content", compile_content, verify_content  },
        { "hexcontent", compile_hexcontent, verify_regcontent },
        { "regcontent", compile_regcontent, verify_regcontent },
        { "ip_opt", compile_ipopt, verify_ipopt               },
        { "tcp_opt", compile_tcpopt, verify_tcpopt            },
       	{ NULL, }
};

static int real_error = 0;
static int rule_count;

static char *get_rule_arg(const char *rule) 
{
        int len = 0;
        char *arg = NULL;
        int i = 0;
        int punct = -1;

        rule_count = 0;
        
        while ( isprint(rule[i]) ) {
                if ( rule[i] == '"' ) {
                        rule_count++;
                        
                        if ( punct == 1 )
                                break;
                        
                        punct = 1;
                        i++;
                        continue;
                }

                if ( isspace(rule[i]) && punct == -1 ) {
                        rule_count++;
                        break;
                }
                
                arg = realloc(arg, ++len * sizeof(char));
                arg[len - 1] = rule[i];
                i++;
        }

        if ( arg == NULL )
                return NULL;
        
        arg = realloc(arg, ++len * sizeof(char));
        arg[len - 1] = '\0';

        return arg;
}

static int rule_handler(DetectPluginFilter_t *plug, const char *rule) 
{
        struct bpf_rule *p;
        int ret = -1;
        char *arg;
        
        for ( p = rules_table; p->rule_name != NULL; ++p ) 
                if ( strncmp(rule, p->rule_name, strlen(p->rule_name)) == 0 ) {
                        plug->name = p->rule_name;
                        plug->func = p->run;
                        ret = strlen(p->rule_name);
                        break;
                }

        if ( ret < 0 ) {
		ret = 0;	
                fprintf(stderr, "parse error: can't understand rule \"%s\"", rule);
                
                while ( rule[ret] != '=' && rule[ret] != '\0' && rule[ret] != ' ' )
                        fprintf(stderr, "%c", rule[ret++]);

                fprintf(stderr, "\".\n");
                
                return -1;
        }

        while ( rule[ret] == ' ' ) ret++;

        if ( rule[ret] == '=' ) {
		ret++;
                
                arg = get_rule_arg(&rule[ret]);
                if ( ! arg ) {
                        fprintf(stderr, "No argument found (rule=%s).\n", rule);
                        return -1;
                }

                ret += strlen(arg) + rule_count;
                
                plug->arg_name = arg;
		if ( p->compile ) {
                	plug->arg = p->compile(arg);
                	if ( plug->arg == -1 ) {
                	        fprintf(stderr, "\t\tthere was an error compiling rule \"%s\" with arg \"%s\".\n",
               	                	p->rule_name, arg);
               	         	free(arg);
                        	return -1;
                	}
		} 
        }

        while ( rule[ret] == ' ' ) ret++;
        
        return ret;
}


static int is_rule_negative(DetectPluginFilter_t *plug,  const char *rule) 
{
        int ret = 0;
        
        if ( rule[ret] == '!' ) {
                ret = 1;
        } else if ( (strncmp(rule, "not", 3) == 0) ) {
                ret = 3;
        }

        if ( ret > 0 )
                plug->expected = -1;
        
        while ( rule[ret] == ' ' ) ret++;

        return ret;
}

static int rule_num = 0;

static int get_operator(DetectPluginFilter_t *plug, const char *rule) 
{
        int ret = 0;
        
        if ( rule_num == 0 )
                return 0;
        
        rule_num = 0;

        if ( (strncmp(rule, "and", 3) == 0) ) {
                plug->and = 1;
                ret = 3;
        } else if ( (strncmp(rule, "&&", 2) == 0) ) {
                plug->and = 1;
                ret = 2;
        } else if ( (strncmp(rule, "or", 2) == 0) ) {
                plug->or = 1;
                ret = 2;
        } else if ( (strncmp(rule, "||", 2) == 0) ) {
                plug->or = 1;
                ret = 2;
        } else {
                real_error = 1;
                fprintf(stderr, "parse error: rule must be separated by \"and\" || \"or\". (rule=%s)\n", rule);
                ret = -1;
        }

        while ( rule[ret] == ' ' ) ret++;

        return ret;
}

static int handle_parenthesis(DetectPluginFilter_t *plug, const char *rule) 
{
        int ret = 0;
        static int rule_id = 0;
        
        if ( rule[ret] == '(' )
                ret = 1;
        else if ( rule[ret] == ')' ) {
                rule_id++;
                ret = 1;
        }
        
        plug->rule_id = rule_id;

        while ( rule[ret] == ' ' ) ret++;

        return ret;
}        


static int i = 0;

static DetectPluginFilter_t *bpf_compile(const char *rule) 
{
        int ret;
        DetectPluginFilter_t *plug;

        plug = calloc(1, sizeof(DetectPluginFilter_t));

        while ( i < strlen(rule) ) {                
                while (rule[i] == ' ' ) i++;

                i += is_rule_negative(plug, &rule[i]);
 
                i += handle_parenthesis(plug, &rule[i]);
                
		ret = get_operator(plug, &rule[i]);
                if ( ret < 0 ) 
                        return NULL;

                i += ret;

                ret = rule_handler(plug, &rule[i]);
                if ( ret < 0 )
                        return NULL;
                
                i += ret;
                rule_num = 1;
                
                return plug;
        }
       
        return NULL;
}

#if 0
static void print_debug(DetectPluginFilter_t *plug) 
{
        printf("\n%d ---> ", plug->rule_id);
        if ( plug->and == 1 ) printf("and ");
        if ( plug->or == 1 ) printf("or ");
        printf("%s %s to be %s\n",
               (plug->expected == 0 ) ? "want" : "don't want", plug->name, plug->arg_name);
}
#endif

DetectPluginFilter_t *DetectPluginFilter_init(const char *rule)
{
        DetectPluginFilter_t *beg = NULL, *cur = NULL, *tmp;
	i = 0;
        rule_num = 0;

        while ( ( tmp = bpf_compile(rule) ) != NULL ) {                
                if ( beg == NULL ) {
                        beg = tmp;
                        cur = tmp;
                } else {
                        cur->next = tmp;
                        cur = cur->next;
                }
        }

        if ( real_error == 1 ) {
                real_error = 0;

                while ( beg != NULL ) {
                        tmp = beg->next;
                        free(beg);
                        beg = tmp;
                }                
                return NULL;
        }   
        return beg;
}

int DetectPluginFilter_run(DetectPluginFilter_t *list) 
{
        int ret;
        register DetectPluginFilter_t *tmp;
        int test_id;

        tmp = list;

  begin:
        while ( tmp != NULL ) {
                test_id = tmp->rule_id;
                
                ret = tmp->func(tmp->arg);
                if ( ret != tmp->expected ) {
                        /*
                         * search "OR" item
                         */
                        for (tmp = tmp->next; tmp != NULL; tmp = tmp->next )
                                if ( tmp->or == 1 && (test_id != tmp->rule_id) )
                                        goto begin;
                        return -1;
                } else if ( ret == tmp->expected ) {
                        /*
                         * Search for "AND" item.
                         */
                        if ( tmp->next && tmp->next->and == 1 && tmp->next->rule_id == test_id) {
                                tmp = tmp->next;
                                goto begin;
                        }
                        return 0;
                }
        }
        return -1;
}













