/*****
*
* Copyright (C) 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdarg.h>

#include "packet.h"
#include "timer.h"
#include "common.h"
#include "alert.h"
#include "alert-prv.h"
#include "plugin-common.h"
#include "rqueue.h"
#include "recycler.h"
#include "plugin-detect.h"
#include "detect-plugins-api.h"
#include "rsend.h"


/*
 * Set date on an alert structure.
 */
__inline__ static void alert_set_date(alert_t *alert) 
{
        time_t t = time(NULL);
        alert_date(alert) = strdup(ctime(&t));
}



/*
 * Function called back when one of the report queue
 * item expire.
 */
static void rqueue_expire(unsigned long ptr) 
{
        rqueue_t *item = (rqueue_t *) ptr;
        
        timer_destroy(&item->timer);
        
        printf("Debug: Flushing report for id=%p, count=%d...\n",
               item, alert_count(&item->report));
        
        rsend_emmit(&item->report);        

        packet_release(alert_packet(&item->report));
        
        free(alert_date(&item->report));
        
        alert_count(&item->report) = 0;
}



/*
 * First time we add this to our report queue.
 */
static void rqueue_first(rqueue_t *item) 
{
        alert_count(&item->report) = 1;
        timer_set_expire(&item->timer, 1);
        timer_set_callback(&item->timer, rqueue_expire);
        timer_set_data(&item->timer, item);       
        timer_init(&item->timer);
}



/*
 * Update one of our report queue item.
 */
__inline__ static void rqueue_update(rqueue_t *item) 
{        
        alert_count(&item->report)++;
        timer_reset(&item->timer);
}



/*
 *
 */
static void rqueue_generic(rqueue_t *item, alert_t *report, Packet_t *packet) 
{
        /*
         * if kind is normal, source IP can't be faked,
         * so we must do one report for one attack...
         * do not queue it.
         */
        if ( report->kind == normal ) {
                alert_count(report) = 1;
                rsend_emmit(report);
                free(alert_date(report));
        } else {
                packet_lock(packet);
                rqueue_first(item);
        }
}



/*
 * Queue a report depending on it's kind.
 */
void plugin_rqueue_report(rqueue_t *item, plugin_generic_t *plugin,
                          Packet_t *packet, rkind_t kind, const char *msg, ...) 
{
        alert_t *report = &item->report;
        va_list va;
        
        if ( report->count > 0 && kind != normal ) {
                rqueue_update(item);
                return;
        }

        va_start(va, msg);
        vsnprintf(alert_message(report), sizeof(alert_message(report)), msg, va);
        va_end(va);
        
        alert_set_date(report);
        alert_packet(report) = packet;
        alert_kind(report) = kind;
        alert_from(report) = plugin->name;
        alert_desc(report) = plugin->desc;

        rqueue_generic(item, report, packet);
}



/*
 *
 */
void prelude_rqueue_report(rqueue_t *item, Packet_t *packet,
                           rkind_t kind, const char *msg, ...) 
{
        alert_t *report = &item->report;
        va_list va;
        
        if ( report->count > 0 && kind != normal ) {
                rqueue_update(item);
                return;
        }

        va_start(va, msg);
        vsnprintf(alert_message(report), sizeof(alert_message(report)), msg, va);
        va_end(va);
        
        alert_set_date(report);
        alert_packet(report) = packet;
        alert_kind(report) = kind;
        alert_from(report) = "Prelude Core";
        alert_desc(report) = "Prelude Core detection engine";

        rqueue_generic(item, report, packet);
}






