/*****
*
* Copyright (C) 1999 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
* 20000207 (fg) : rewrote SendSavedReport to use sprintf()
*
*****/

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <rpc/xdr.h>
#include <signal.h>

#include "packet.h"
#include "alert.h"
#include "alert-prv.h"
#include "rxdr.h"
#include "pconfig.h"
#include "rsend.h"
#include "rsave.h"
#include "common.h"
#include "print.h"
#include "timer.h"
#include "auth.h"

#define UNIX_SOCK "/dev/prelude_socket"

static int saved_report = 0;
static int sock;
static XDR xdrs;
extern Pconfig_t config;

static int _write_xdr(void *handle, char *buf, int len);



static void sigpipe_handler(int sig) 
{
        p_print(stdout, "Caught signal %d\n", sig);
}



/*
 * Connect to the specified address in a generic manner
 * (can be Unix or Inet ).
 */
static int generic_connect(struct sockaddr *addr, size_t addr_size) 
{
        int ret;
        
	ret = connect(sock, addr, addr_size);
	if ( ret < 0 ) {
                err("error connecting to unix socket");
                return -1;
	}
        
        ret = auth_client(sock);
        if ( ret != 0 ) {
                err("authentication to report server failed.");
                return -1;
        }

        ret = fcntl(sock, F_SETFL, O_NONBLOCK);
        if ( ret < 0 ) {
                perror("fcntl");
                return -1;
        }
        
        return 0;
}



/*
 * Setup an Unix connection.
 */
static int unix_connect(void)
{
	struct sockaddr_un addr;
        int ret;
        
	p_print(stdout, "\tConnecting to Unix prelude report server.\n");

	sock = socket(AF_UNIX, SOCK_STREAM, 0);
	if (sock < 0) {
                err("Error opening unix socket");
		return -1;
	}
        
	addr.sun_family = AF_UNIX;
	strcpy(addr.sun_path, UNIX_SOCK);

        ret = generic_connect((struct sockaddr *)&addr, sizeof(addr));
        if ( ret < 0 ) {
                err("Error connecting to report server");
                close(sock); sock = -1;
                return -1;
        }

	return 0;
}



/*
 * Setup an Inet connection.
 */
static int inet_connect(const char *server, unsigned int port)
{
	struct sockaddr_in addr;
        int ret;
                
	p_print(stdout, "Connecting to Tcp prelude report server.\n");

	sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sock < 0) {
		err("Error opening inet socket");
		return -1;
	}

	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = inet_addr(server);

        ret = generic_connect((struct sockaddr *)&addr, sizeof(addr));
        if ( ret < 0 ) {
                err("Error connecting to %s", server);
                close(sock); sock = -1;
                return -1;
        }
        
	return 0;
}



static int do_connect(const char *server, int port) 
{
        int ret;
        
        ret = strcmp(server, "unix");
	if (ret == 0)
		return unix_connect();
	else
		return inet_connect(server, port);
}



/*
 * function called back when a report timer expire.
 */
static void expire(unsigned long data) 
{
        int ret;
        timer_t *timer = (timer_t *) data;

        ret = do_connect(config.addr, config.port);
        if ( ret < 0 ) {
                err("can't reconnect to report server, %d report in queue",
                    saved_report);
                
                timer_set_expire(timer, timer_expire(timer) * 2);
                timer_reset(timer);
                return;
        }

        timer_destroy(timer);
        backout_report((void *)&sock, _write_xdr);
}




__inline__ static void report_timer_set(void) 
{
        static timer_t timer;
        timer_t *t = &timer;
        
        timer_set_callback(t, expire);
        timer_set_data(t, t);
        timer_set_expire(t, 10);

        timer_init(t);
}



static int _write(int fd, const char *buf, const int len) 
{
        int ret, i = 10;

        do {
                ret = write(sock, buf, len);
                if ( ret < 0 && errno != EAGAIN ) {
                        err("write");
                        return -1;
                }
        } while ( i-- && ret < 0 );

        printf("written %d bytes\n", ret);
        
        return ret;
}



static int _write_xdr(void *handle, char *buf, int len) 
{
        int ret;
        char *bkp = buf;
        
        if ( sock < 0 ) {
                backup_report(bkp, len);
                return len;
        }

        ret = _write(sock, buf, len);
        if ( ret < 0 ) {
                close(sock); sock = -1;
                backup_report(bkp, len);
                report_timer_set();
                return -1;
        }
        
        return len;
}



/*
 * Emmit a new report.
 */
void rsend_emmit(alert_t *report)
{
        int ret;
        
        ret = __rxdr_encode(&xdrs, report);
        if ( ret < 0 ) {
                err("An error occured while converting report into XDR data");
                return;
        }
}



/*
 * Init the 'report send' module :
 * - connect to the report server.
 */
int rsend_init(const char *server, unsigned int port)
{
	int ret;

        signal(SIGPIPE, sigpipe_handler);
        
        ret = do_connect(server, port);
        if ( ret < 0 )
                return -1;

        xdrrec_create(&xdrs, 0, 0, NULL, NULL, _write_xdr);
        xdrs.x_op = XDR_ENCODE;
        
	return 0;
}










