/*****
*
* Copyright (C) 1999 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>

/*
 * extract.h stolen from tcpdump source
 */
#include "extract.h"
#include "common.h"
#include "packet.h"

#include "proto.h"
#include "alert.h"
#include "tcpip_options.h"

/*
 * From rfc793 :
 * Options may occupy space at the end of the TCP header and are a
 * multiple of 8 bits in length.  All options are included in the
 * checksum.  An option may begin on any octet boundary.  There are two
 * cases for the format of an option:
 *
 *  Case 1:  A single octet of option-kind.
 *
 *  Case 2:  An octet of option-kind, an octet of option-length, and
 *           the actual option-data octets.
 *
 * The option-length counts the two octets of option-kind and
 * option-length as well as the option-data octets.
 *
 * Note that the list of options may be shorter than the data offset
 * field might imply.  The content of the header beyond the
 * End-of-Option option must be header padding (i.e., zero).
 *
 * A TCP must implement all options.
 */

/*
 * This function parse the tcp options,
 * to see if they are valid, it will return
 * an integer delimiting the the valid options found in the options buffer.
 */

extern int depth;

int VerifyOptions(const unsigned char *opt_buf, unsigned int optlen,
                  const unsigned char *snapend, Packet_t *packet,
                  const char *proto)
{
	int datalen, len, kind, size = 0;
        
        while (optlen > 0 && size < __MAX_OPTS_LEN) {
                /*
		 * Take the option kind value.
		 */
		kind = *opt_buf++;
                
		/*
                 * If option kind is eol or nop
		 * len is 1.
                 * eol & nop have the same value in IP & TCP.
		 */
                if (kind == __IPOPT_NOP || kind == __IPOPT_EOL) {
                        len = 1;
                        optlen--;
                        datalen = 0;
                        size++;
                        continue;
                } else {
                        /*
                         * We register our option len &
                         * we advance our pointer onto the next byte.
                         */
                        len = *opt_buf++;
                        size += 2;
                        
                        /*
                         * Option len should not be < 2 bytes
                         * ( minimum option size except for eol & nop )
                         * and should not be > than our total options len.
                         */
                        if ( len < 2 ) {
                                prelude_do_report(packet, maybe_faked, RID_OPTS_LEN,
                                                  "%s options not valid : len (%d) should be >= 2.\n\n"
                                                  "This could be issued in order to make your operating"
                                                  " system\n"
                                                  "to handle this packet in an unpredictable way\n"\
                                                  "or as a DOS attack.",
                                                  proto, len);
                                return size;
                        }
 
                        if ( len > optlen ) {
                                prelude_do_report(packet, maybe_faked, RID_OPTS_TOTLEN,
                                                  "%s options not valid : len (%d) is > total options len(%d)\n\n"\
                                                  "This could be issued in order to make your operating system\n"\
                                                  "to handle this packet in an unpredictable way\n"\
                                                  "or as a DOS attack.",
                                                  proto, len, optlen);
                                return size;
                        }
                        
                        /*
                         * We verify that our options data byte
                         * doesn't point outside our buffer space
                         */
                        datalen = len - 2;
                        if ( opt_buf + datalen > snapend ) {
                                prelude_do_report(packet, maybe_faked, RID_OPTS_TRUNC,
                                                  "%s options header seem to be truncated,\n"\
                                                  "options data try to make us point on an address\n"\
                                                  "which is > than our packet data space\n"\
                                                  "(option data addr at %p > end of packet at %p)\n\n"\
                                                  "This could be issued in order to make your operating system\n"\
                                                  "to handle this packet in an unpredictable way\n"\
                                                  "or as a DOS attack.",
                                                  proto, opt_buf + datalen, snapend);
                                return size;
                        }
                        
                        size += datalen;
                }

                opt_buf += datalen;
		optlen -= (datalen + 2);
	}
	return size;
}
















